@@ -4,7 +4,7 @@ use crate::explorer;
 use crate::input::Key;
 use crate::lua;
 use crate::permissions::Permissions;
-use crate::runner;
+use crate::runner::Runner;
 use crate::ui::Layout;
 use anyhow::{bail, Result};
 use chrono::{DateTime, Local};
@@ -366,6 +366,7 @@ impl Ord for Node {
         other.relative_path.cmp(&self.relative_path)
     }
 }
+
 impl PartialOrd for Node {
     fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
         Some(self.cmp(other))
@@ -418,7 +419,8 @@ impl DirectoryBuffer {
 
 #[derive(Debug, Clone, Eq, PartialEq, Serialize, Deserialize)]
 pub enum InternalMsg {
-    AddDirectory(String, DirectoryBuffer),
+    AddLastFocus(String, Option<String>),
+    SetDirectory(DirectoryBuffer),
     HandleKey(Key),
 }
 
@@ -1272,6 +1274,17 @@ pub enum ExternalMsg {
     /// Toggle mouse
     ToggleMouse,
 
+    /// Start piping the focused path to the given fifo path
+    ///
+    /// **Example:** `StartFifo: /tmp/xplr.fifo`
+    StartFifo(String),
+
+    /// Close the active fifo and stop piping.
+    StopFifo,
+
+    /// Toggle betwen {Start|Stop}Fifo
+    ToggleFifo(String),
+
     /// Log information message.
     ///
     /// **Example:** `LogInfo: launching satellite`
@@ -1295,7 +1308,16 @@ pub enum ExternalMsg {
     /// Quit with returncode zero (success).
     Quit,
 
-    /// Print selected paths if it's not empty, else, print the focused node's path.
+    /// Print $PWD and quit.
+    PrintPwdAndQuit,
+
+    /// Print the path under focus and quit. It can be empty string if there's nothing to focus.
+    PrintFocusPathAndQuit,
+
+    /// Print the selected paths and quit. It can be empty is no path is selected.
+    PrintSelectionAndQuit,
+
+    /// Print the selected paths if it's not empty, else, print the focused node's path.
     PrintResultAndQuit,
 
     /// Print the state of application in YAML format. Helpful for debugging or generating
@@ -1352,8 +1374,6 @@ pub enum MsgOut {
     Refresh,
     ClearScreen,
     Quit,
-    PrintResultAndQuit,
-    PrintAppStateAndQuit,
     Debug(String),
     Call(Command),
     CallSilently(Command),
@@ -1363,6 +1383,14 @@ pub enum MsgOut {
     EnableMouse,
     DisableMouse,
     ToggleMouse,
+    StartFifo(String),
+    StopFifo,
+    ToggleFifo(String),
+    PrintPwdAndQuit,
+    PrintFocusPathAndQuit,
+    PrintSelectionAndQuit,
+    PrintResultAndQuit,
+    PrintAppStateAndQuit,
 }
 
 #[derive(Debug, Clone, Eq, PartialEq, Serialize, Deserialize)]
@@ -1490,7 +1518,8 @@ pub struct App {
     pub version: String,
     pub config: Config,
     pub pwd: String,
-    pub directory_buffers: HashMap<String, DirectoryBuffer>,
+    pub directory_buffer: Option<DirectoryBuffer>,
+    pub last_focus: HashMap<String, Option<String>>,
     pub selection: IndexSet<Node>,
     pub msg_out: VecDeque<MsgOut>,
     pub mode: Mode,
@@ -1507,17 +1536,19 @@ pub struct App {
 }
 
 impl App {
-    pub fn create(pwd: PathBuf, lua: &mlua::Lua) -> Result<Self> {
+    pub fn create(pwd: PathBuf, lua: &mlua::Lua, config_file: Option<PathBuf>) -> Result<Self> {
         let config = lua::init(lua)?;
 
-        let config_dir = dirs::config_dir()
-            .unwrap_or_else(|| PathBuf::from("."))
-            .join("xplr");
+        let config_file = if let Some(path) = config_file {
+            path
+        } else if let Some(dir) = dirs::home_dir() {
+            dir.join(".config").join("xplr").join("init.lua")
+        } else {
+            PathBuf::from("/").join("etc").join("xplr").join("init.lua")
+        };
 
-        let lua_script_file = config_dir.join("init.lua");
-
-        let (config, load_err) = if lua_script_file.exists() {
-            match lua::extend(lua, &lua_script_file.to_string_lossy().to_string()) {
+        let (config, load_err) = if config_file.exists() {
+            match lua::extend(lua, &config_file.to_string_lossy().to_string()) {
                 Ok(c) => (c, None),
                 Err(e) => (config, Some(e.to_string())),
             }
@@ -1553,7 +1584,7 @@ impl App {
 
         let pid = std::process::id();
         let session_path = dirs::runtime_dir()
-            .unwrap_or_else(|| "/tmp".into())
+            .unwrap_or_else(env::temp_dir)
             .join("xplr")
             .join("session")
             .join(&pid.to_string())
@@ -1579,7 +1610,8 @@ impl App {
             version: VERSION.to_string(),
             config,
             pwd,
-            directory_buffers: Default::default(),
+            directory_buffer: Default::default(),
+            last_focus: Default::default(),
             selection: Default::default(),
             msg_out: Default::default(),
             mode,
@@ -1595,6 +1627,8 @@ impl App {
             last_modes: Default::default(),
         };
 
+        fs::create_dir_all(app.session_path())?;
+
         if let Some(err) = load_err {
             app.log_error(err)
         } else {
@@ -1627,7 +1661,10 @@ impl App {
 
     fn handle_internal(self, msg: InternalMsg) -> Result<Self> {
         match msg {
-            InternalMsg::AddDirectory(parent, dir) => self.add_directory(parent, dir),
+            InternalMsg::SetDirectory(dir) => self.set_directory(dir),
+            InternalMsg::AddLastFocus(parent, focus_path) => {
+                self.add_last_focus(parent, focus_path)
+            }
             InternalMsg::HandleKey(key) => self.handle_key(key),
         }
     }
@@ -1718,11 +1755,17 @@ impl App {
                 ExternalMsg::EnableMouse => self.enable_mouse(),
                 ExternalMsg::DisableMouse => self.disable_mouse(),
                 ExternalMsg::ToggleMouse => self.toggle_mouse(),
+                ExternalMsg::StartFifo(f) => self.start_fifo(f),
+                ExternalMsg::StopFifo => self.stop_fifo(),
+                ExternalMsg::ToggleFifo(f) => self.toggle_fifo(f),
                 ExternalMsg::LogInfo(l) => self.log_info(l),
                 ExternalMsg::LogSuccess(l) => self.log_success(l),
                 ExternalMsg::LogWarning(l) => self.log_warning(l),
                 ExternalMsg::LogError(l) => self.log_error(l),
                 ExternalMsg::Quit => self.quit(),
+                ExternalMsg::PrintPwdAndQuit => self.print_pwd_and_quit(),
+                ExternalMsg::PrintFocusPathAndQuit => self.print_focus_path_and_quit(),
+                ExternalMsg::PrintSelectionAndQuit => self.print_selection_and_quit(),
                 ExternalMsg::PrintResultAndQuit => self.print_result_and_quit(),
                 ExternalMsg::PrintAppStateAndQuit => self.print_app_state_and_quit(),
                 ExternalMsg::Debug(path) => self.debug(path),
@@ -1756,12 +1799,16 @@ impl App {
             })
             .or_else(|| default.map(|a| a.messages().clone()))
             .unwrap_or_else(|| {
-                vec![
-                    ExternalMsg::SwitchModeBuiltin("recover".into()),
-                    ExternalMsg::LogWarning(
-                        "Key map not found. Let's calm down, escape, and try again.".into(),
-                    ),
-                ]
+                if self.config().general().disable_recover_mode() {
+                    vec![ExternalMsg::LogWarning("Key map not found.".into())]
+                } else {
+                    vec![
+                        ExternalMsg::SwitchModeBuiltin("recover".into()),
+                        ExternalMsg::LogWarning(
+                            "Key map not found. Let's calm down, escape, and try again.".into(),
+                        ),
+                    ]
+                }
             });
 
         for msg in msgs {
@@ -1771,14 +1818,17 @@ impl App {
         Ok(self)
     }
 
-    pub fn explore_pwd(self) -> Result<Self> {
+    pub fn explore_pwd(mut self) -> Result<Self> {
+        let focus = self.last_focus.get(self.pwd()).cloned().unwrap_or(None);
+        let pwd = self.pwd().clone();
+        self = self.add_last_focus(pwd, focus.clone())?;
         let dir = explorer::explore_sync(
             self.explorer_config().clone(),
             self.pwd().into(),
-            self.focused_node().map(|n| n.absolute_path().into()),
+            focus.map(PathBuf::from),
             self.directory_buffer().map(|d| d.focus()).unwrap_or(0),
         )?;
-        self.add_directory(dir.parent().clone(), dir)
+        self.set_directory(dir)
     }
 
     fn explore_pwd_async(mut self) -> Result<Self> {
@@ -1839,7 +1889,11 @@ impl App {
 
     fn focus_previous(mut self) -> Result<Self> {
         if let Some(dir) = self.directory_buffer_mut() {
-            dir.focus = dir.focus.max(1) - 1;
+            dir.focus = if dir.focus == 0 {
+                dir.total.max(1) - 1
+            } else {
+                dir.focus.max(1) - 1
+            };
         };
         Ok(self)
     }
@@ -1869,7 +1923,11 @@ impl App {
 
     fn focus_next(mut self) -> Result<Self> {
         if let Some(dir) = self.directory_buffer_mut() {
-            dir.focus = (dir.focus + 1).min(dir.total.max(1) - 1);
+            dir.focus = if (dir.focus + 1) == dir.total {
+                0
+            } else {
+                dir.focus + 1
+            }
         };
         Ok(self)
     }
@@ -1909,13 +1967,21 @@ impl App {
     }
 
     fn change_directory(mut self, dir: &str, save_history: bool) -> Result<Self> {
-        match env::set_current_dir(dir) {
+        let mut dir = PathBuf::from(dir);
+        if dir.is_relative() {
+            dir = PathBuf::from(self.pwd()).join(dir);
+        }
+
+        match env::set_current_dir(&dir) {
             Ok(()) => {
-                self.pwd = dir.to_owned();
+                let pwd = self.pwd().clone();
+                let focus = self.focused_node().map(|n| n.relative_path().clone());
+                self = self.add_last_focus(pwd, focus)?;
+                self.pwd = dir.to_string_lossy().to_string();
                 if save_history {
                     self.history = self.history.push(format!("{}/", self.pwd));
                 }
-                self.explore_pwd_async()
+                self.explore_pwd()
             }
             Err(e) => self.log_error(e.to_string()),
         }
@@ -2078,7 +2144,10 @@ impl App {
     }
 
     pub fn focus_path(self, path: &str, save_history: bool) -> Result<Self> {
-        let pathbuf = PathBuf::from(path);
+        let mut pathbuf = PathBuf::from(path);
+        if pathbuf.is_relative() {
+            pathbuf = PathBuf::from(self.pwd()).join(pathbuf);
+        }
         if let Some(parent) = pathbuf.parent() {
             if let Some(filename) = pathbuf.file_name() {
                 self.change_directory(&parent.to_string_lossy().to_string(), false)?
@@ -2224,8 +2293,19 @@ impl App {
         })
     }
 
-    pub fn add_directory(mut self, parent: String, dir: DirectoryBuffer) -> Result<Self> {
-        self.directory_buffers.insert(parent, dir);
+    pub fn set_directory(mut self, dir: DirectoryBuffer) -> Result<Self> {
+        self = self.add_last_focus(
+            dir.parent.clone(),
+            dir.focused_node().map(|n| n.relative_path().clone()),
+        )?;
+        if dir.parent == self.pwd {
+            self.directory_buffer = Some(dir);
+        }
+        Ok(self)
+    }
+
+    pub fn add_last_focus(mut self, parent: String, focused_path: Option<String>) -> Result<Self> {
+        self.last_focus.insert(parent, focused_path);
         Ok(self)
     }
 
@@ -2239,7 +2319,10 @@ impl App {
     }
 
     fn select_path(mut self, path: String) -> Result<Self> {
-        let path = PathBuf::from(path);
+        let mut path = PathBuf::from(path);
+        if path.is_relative() {
+            path = PathBuf::from(self.pwd()).join(path);
+        }
         let parent = path.parent().map(|p| p.to_string_lossy().to_string());
         let filename = path.file_name().map(|p| p.to_string_lossy().to_string());
         if let (Some(p), Some(n)) = (parent, filename) {
@@ -2268,7 +2351,12 @@ impl App {
     }
 
     fn un_select_path(mut self, path: String) -> Result<Self> {
-        self.selection.retain(|n| n.absolute_path != path);
+        let mut pathbuf = PathBuf::from(path);
+        if pathbuf.is_relative() {
+            pathbuf = PathBuf::from(self.pwd()).join(pathbuf);
+        }
+        self.selection
+            .retain(|n| PathBuf::from(&n.absolute_path) != pathbuf);
         Ok(self)
     }
 
@@ -2282,15 +2370,12 @@ impl App {
         Ok(self)
     }
 
-    fn toggle_selection(mut self) -> Result<Self> {
-        if let Some(n) = self.focused_node() {
-            if self.selection().contains(n) {
-                self = self.un_select()?;
-            } else {
-                self = self.select()?;
-            }
+    fn toggle_selection(self) -> Result<Self> {
+        if let Some(p) = self.focused_node().map(|n| n.absolute_path().clone()) {
+            self.toggle_selection_by_path(p)
+        } else {
+            Ok(self)
         }
-        Ok(self)
     }
 
     fn toggle_select_all(self) -> Result<Self> {
@@ -2306,10 +2391,18 @@ impl App {
     }
 
     fn toggle_selection_by_path(self, path: String) -> Result<Self> {
-        if self.selection.iter().any(|n| n.absolute_path == path) {
-            self.select_path(path)
-        } else {
+        let mut pathbuf = PathBuf::from(&path);
+        if pathbuf.is_relative() {
+            pathbuf = PathBuf::from(self.pwd()).join(pathbuf);
+        }
+        if self
+            .selection
+            .iter()
+            .any(|n| PathBuf::from(&n.absolute_path) == pathbuf)
+        {
             self.un_select_path(path)
+        } else {
+            self.select_path(path)
         }
     }
 
@@ -2448,6 +2541,21 @@ impl App {
         Ok(self)
     }
 
+    fn start_fifo(mut self, path: String) -> Result<Self> {
+        self.msg_out.push_back(MsgOut::StartFifo(path));
+        Ok(self)
+    }
+
+    fn stop_fifo(mut self) -> Result<Self> {
+        self.msg_out.push_back(MsgOut::StopFifo);
+        Ok(self)
+    }
+
+    fn toggle_fifo(mut self, path: String) -> Result<Self> {
+        self.msg_out.push_back(MsgOut::ToggleFifo(path));
+        Ok(self)
+    }
+
     pub fn log_info(mut self, message: String) -> Result<Self> {
         self.logs_hidden = false;
         self.logs.push(Log::new(LogLevel::Info, message));
@@ -2477,6 +2585,21 @@ impl App {
         Ok(self)
     }
 
+    fn print_pwd_and_quit(mut self) -> Result<Self> {
+        self.msg_out.push_back(MsgOut::PrintPwdAndQuit);
+        Ok(self)
+    }
+
+    fn print_focus_path_and_quit(mut self) -> Result<Self> {
+        self.msg_out.push_back(MsgOut::PrintFocusPathAndQuit);
+        Ok(self)
+    }
+
+    fn print_selection_and_quit(mut self) -> Result<Self> {
+        self.msg_out.push_back(MsgOut::PrintSelectionAndQuit);
+        Ok(self)
+    }
+
     fn print_result_and_quit(mut self) -> Result<Self> {
         self.msg_out.push_back(MsgOut::PrintResultAndQuit);
         Ok(self)
@@ -2493,7 +2616,7 @@ impl App {
     }
 
     fn directory_buffer_mut(&mut self) -> Option<&mut DirectoryBuffer> {
-        self.directory_buffers.get_mut(&self.pwd)
+        self.directory_buffer.as_mut()
     }
 
     /// Get a reference to the app's pwd.
@@ -2503,7 +2626,7 @@ impl App {
 
     /// Get a reference to the app's current directory buffer.
     pub fn directory_buffer(&self) -> Option<&DirectoryBuffer> {
-        self.directory_buffers.get(&self.pwd)
+        self.directory_buffer.as_ref()
     }
 
     /// Get a reference to the app's config.
@@ -2529,11 +2652,6 @@ impl App {
         format!("{}\n", &self.mode.name())
     }
 
-    /// Get a reference to the app's directory buffers.
-    pub fn directory_buffers(&self) -> &HashMap<String, DirectoryBuffer> {
-        &self.directory_buffers
-    }
-
     /// Get a reference to the app's input buffer.
     pub fn input_buffer(&self) -> Option<String> {
         self.input_buffer.clone()
@@ -2684,70 +2802,27 @@ impl App {
     }
 
     pub fn write_pipes(&self) -> Result<()> {
-        // TODO optimize and test
-
         fs::create_dir_all(self.pipe().path())?;
-
         fs::write(&self.pipe().msg_in, "")?;
 
         let selection_str = self.selection_str();
-        // if last_app
-        //     .map(|a| a.selection_str() != selection_str)
-        //     .unwrap_or(true)
-        // {
         fs::write(&self.pipe().selection_out, selection_str)?;
-        // };
 
         let history_str = self.history_str();
-        // if last_app
-        //     .map(|a| a.history_str() != history_str)
-        //     .unwrap_or(true)
-        // {
         fs::write(&self.pipe().history_out, history_str)?;
-        // };
 
         let directory_nodes_str = self.directory_nodes_str();
-        // if last_app
-        //     .map(|a| a.directory_nodes_str() != directory_nodes_str)
-        //     .unwrap_or(true)
-        // {
         fs::write(&self.pipe().directory_nodes_out, directory_nodes_str)?;
-        // };
 
-        // if last_app
-        //     .map(|a| a.logs().len() != self.logs().len())
-        //     .unwrap_or(true)
-        // {
-        // let new_logs = self
-        //     .logs()
-        //     .iter()
-        //     .skip(last_app.map(|a| a.logs().len()).unwrap_or(0))
-        //     .map(|l| format!("{}\n", l))
-        //     .collect::<Vec<String>>()
-        //     .join("");
-
-        // let mut file = fs::OpenOptions::new()
-        //     .append(true)
-        //     .open(&self.pipe().logs_out)?;
-        // file.write_all(new_logs.as_bytes())?;
-        // };
         let logs_str = self.logs_str();
         fs::write(&self.pipe().logs_out, logs_str)?;
 
         let result_str = self.result_str();
-        // if last_app
-        //     .map(|a| a.result_str() != result_str)
-        //     .unwrap_or(true)
-        // {
         fs::write(&self.pipe().result_out, result_str)?;
-        // };
+
         let global_help_menu_str = self.global_help_menu_str();
-        // if last_app
-        //     .map(|a| a.global_help_menu_str() != global_help_menu_str)
-        //     .unwrap_or(true)
-        // {
         fs::write(&self.pipe().global_help_menu_out, global_help_menu_str)?;
-        // };
+
         Ok(())
     }
 
@@ -2796,15 +2871,8 @@ impl App {
             last_modes: self.last_modes.clone(),
         }
     }
-
-    pub fn run(self, focused_path: Option<PathBuf>, lua: &mlua::Lua) -> Result<Option<String>> {
-        runner::run(self, focused_path, lua)
-    }
 }
 
-/// Run xplr TUI
-pub fn run(pwd: PathBuf, focused_path: Option<PathBuf>) -> Result<Option<String>> {
-    let lua = mlua::Lua::new();
-    let app = App::create(pwd, &lua)?;
-    app.run(focused_path, &lua)
+pub fn runner(path: Option<PathBuf>) -> Result<Runner> {
+    Runner::new(path)
 }
