@@ -18,15 +18,21 @@ use std::io::Write;
 use std::path::PathBuf;
 use std::process::{Command, ExitStatus, Stdio};
 use std::sync::mpsc;
-use termion::get_tty;
 use tui::backend::CrosstermBackend;
 use tui::Terminal;
 
+pub fn get_tty() -> io::Result<fs::File> {
+    fs::OpenOptions::new()
+        .read(true)
+        .write(true)
+        .open("/dev/tty")
+}
+
 fn call_lua(
     app: &app::App,
     lua: &mlua::Lua,
     func: &str,
-    silent: bool,
+    _silent: bool,
 ) -> Result<Option<Vec<app::ExternalMsg>>> {
     let _focus_index = app
         .directory_buffer()
@@ -34,12 +40,6 @@ fn call_lua(
         .unwrap_or_default()
         .to_string();
 
-    let (_, _, _) = if silent {
-        (Stdio::null(), Stdio::null(), Stdio::null())
-    } else {
-        (get_tty()?.into(), get_tty()?.into(), get_tty()?.into())
-    };
-
     let arg = app.to_lua_arg();
     let arg = lua.to_value(&arg)?;
     lua::call(lua, func, arg)
@@ -106,12 +106,15 @@ pub struct Runner {
 
 impl Runner {
     pub(crate) fn new(path: Option<PathBuf>) -> Result<Self> {
-        let mut pwd = path.unwrap_or_else(|| ".".into()).canonicalize()?;
+        let basedir = std::env::current_dir()?;
+        let mut pwd = path
+            .map(|p| if p.is_relative() { basedir.join(p) } else { p })
+            .unwrap_or_else(|| basedir.clone());
         let mut focused_path = None;
 
         if pwd.is_file() {
             focused_path = pwd.file_name().map(|p| p.into());
-            pwd = pwd.parent().map(|p| p.into()).unwrap_or_else(|| ".".into());
+            pwd = pwd.parent().map(|p| p.into()).unwrap_or(basedir);
         }
 
         Ok(Self {
@@ -139,7 +142,8 @@ impl Runner {
     }
 
     pub fn run(self) -> Result<Option<String>> {
-        let lua = mlua::Lua::new();
+        // Why unsafe? See https://github.com/sayanarijit/xplr/issues/309
+        let lua = unsafe { mlua::Lua::unsafe_new() };
         let mut app = app::App::create(self.pwd, &lua, self.config)?;
         app.config.general.set_read_only(self.read_only);
 
@@ -182,7 +186,7 @@ impl Runner {
         execute!(stdout, term::EnterAlternateScreen)?;
 
         let mut fifo: Option<fs::File> = if let Some(path) = app.config().general().start_fifo() {
-            match start_fifo(&path, &app.focused_node_str()) {
+            match start_fifo(path, &app.focused_node_str()) {
                 Ok(file) => Some(file),
                 Err(e) => {
                     app = app.log_error(e.to_string())?;
@@ -562,7 +566,7 @@ impl Runner {
         term::disable_raw_mode()?;
         terminal.show_cursor()?;
 
-        fs::remove_dir(session_path)?;
+        fs::remove_dir_all(session_path)?;
 
         result
     }
